#include "SREnemySprite.h"
#include "SRAmmoSprite.h"
#include "SRGameObject.h"
#include "SRSoundManager.h"
#include "SRBoomSprite.h"

//#include <iostream>

SREnemySprite::SREnemySprite( int startingID, SRObject *command ) : SRIconSprite(0),
    fIconSwitchTimer(kSwitchCountdown), fIsIconOne(true), fCanFire(false)
{
    char type[5];
    *((int*)type) = Type();
    type[4] = 0;
    //cout << "Object of type " << type << " allocated with address " << this << endl;
    fParent = command;
    
    // Load icons
    fIcon1 = GetCIcon( startingID + kIconOneID );
    fIcon2 = GetCIcon( startingID + kIconTwoID );
    //fAmmoIcon = GetCIcon( kFireIcon );
    
    // create a list for ammo
    //fAmmoList = new SRObjectList;
    
    fX = 64;
    fY = 64;
    
    fWidth = fHeight = 20;
}

SREnemySprite::~SREnemySprite()
{
    // dump the icons and the ammo list
    DisposeCIcon(fIcon1);
    DisposeCIcon(fIcon2);
    
    // TODO: this will destroy the ammo itself...
    //       need to put this list elsewhere, for dealing with afterwards...
    //delete fAmmoList;
}
    
void SREnemySprite::Think()
{
    //fAmmoList->clean();
 
    // decide what clothes to wear
    if( fIconSwitchTimer > 0 ) fIconSwitchTimer--;
    else
    {
        fIconSwitchTimer = kSwitchCountdown;
        if(fIsIconOne)
            fCurrentIcon = fIcon1;
        else
            fCurrentIcon = fIcon2;
        
        fIsIconOne = !fIsIconOne;
    }
    
    // if we can fire (no one's in the way), see if we will
    if(fCanFire)
    {
        short randomNum = Random();
    
        if(randomNum < 0) randomNum = -randomNum;
    
        float percent = (float)randomNum / 32767.0;
    
        // if we fall under the threshold, we let a missile loose
        if(percent < kFireProbability)
        {
            SRAmmoSprite *ammo = new SRAmmoSprite( GetCIcon(kFireIcon), this );
            ammo->setX( fX );
            ammo->setY( fY + kFireDistance );
            ammo->pushY( kFireForce );
            //fAmmoList->add(ammo);
        }
    }
}

bool SREnemySprite::Hit( SRObject *other )
{
    // if what hit us is enemy fire, we die...
    if( other->Type() == kPlayerAmmoType || other->Type() == kGenericAmmoType )
    {
        SRSoundManager::PlaySoundEffect(SRSoundManager::SFXboom);
        fParent->ChildDied(this);
        SRGameObject::instance->Scored(kScoreValue);
        new SRBoomSprite(this);
        return true;
    }
    
    return false;
}

void SREnemySprite::ChildDied( SRObject */*ammo*/ )
{
    // when ammo dies, we remove it...
    //fAmmoList->remove(ammo);
}