#include "SRGameObject.h"
#include "SRPlayerSprite.h"
#include "SREnemy1Sprite.h"
#include "SREnemyCommand.h"
#include "SRStar.h"
#include "WindowDefs.h"
#include "SRSoundManager.h"

#include "FileWrapper.h"
#include "HFileUtils.h"
#include <stdio.h>
#include <string.h>
#include <exception>

FileWrapper *dlog = 0;

//#include <iostream>

bool SRGameObject::fPause = false;
SRGameObject *SRGameObject::instance = 0;

namespace
{
    SRPlayerSprite *player;

// debugging formation - see SREnemyCommand.h for details of format
const char formation[] = { 5,
                           0, 20, -1, 1,
                           -20, -20, -1, 1,
                           20, -20, -1, 1,
                           -40, 0, -1, 1,
                           40, 0, -1, 1 };
                           
const char blockFormation[] = { 12,
                                -48, 32, -1, 1,
                                -16, 32, -1, 1,
                                16, 32, -1, 1,
                                48, 32, -1, 1,
                                -48, 0, 0, 1,
                                -16, 0, 1, 1,
                                16, 0, 2, 1,
                                48, 0, 3, 1,
                                -48, -32, 4, 1,
                                -16, -32, 5, 1,
                                16, -32, 6, 1,
                                48, -32, 7, 1 };
}

SRGameObject::SRGameObject() : fEnemyCommand(0), fScore(0), fTimesDied(0), fList(0)
{
    instance = this;
    
    /*FSSpec spec = GetCurrentApplicationFSSpec();
    sprintf((char *)&(spec.name[1]), "debug.txt" );
    spec.name[0] = strlen((char *)&(spec.name[1]));

    if(!dlog)
    {
        dlog = new FileWrapper( &spec, FileWrapper::kWriteOnly, true, 'TEXT', 'R*ch' );
        if(!dlog->open()) SysBeep(1);
    }*/

    CreateStarField();

    fList = new SRObjectList(false, true);

    player = new SRPlayerSprite(128);
    // Testing only
    //new SREnemy1Sprite(140);
    
    NewGame();

    SRSoundManager::SetMusic(SRSoundManager::MUSaggravated);
    SRSoundManager::StartMusic();
}

SRGameObject::~SRGameObject()
{
    dlog->close();
    delete fList;
    delete fStarField;
}

void SRGameObject::NewGame()
{
//    cout << "Starting new game." << endl;
    // Initialize new enemies to play against
    SRObjectList::Iterator *it = fList->NewIterator();
    SRObject *currentObject;
    
    while( (currentObject = it->next()) != 0 )
    {
        if( currentObject->Type() != kPlayerType )
        {
            fList->remove(currentObject);
        }
    }
    
    fList->clean();
    
    fScore = 0;
    fTimesDied = 0;

    fEnemyCommand = new SREnemyCommand(formation);
    
    player->ResetPosition();
}

void SRGameObject::LevelUp()
{
    // Initialize next level
    SRObjectList::Iterator *it = fList->NewIterator();
    SRObject *currentObject;
    
    while( (currentObject = it->next()) != 0 )
    {
        if( currentObject->Type() != kPlayerType && 
            currentObject->Type() != kEnemyAmmoType &&
            currentObject->Type() != kExplosionType )
        {
            fList->remove(currentObject);
        }
    }
    
    fList->clean();

    fEnemyCommand = new SREnemyCommand(blockFormation);
}

void SRGameObject::GameOver()
{
    SysBeep(1);
    NewGame();
}

// The following three methods delegate their functions to the 
// objects in the object list
void SRGameObject::Draw()
{
    SRObject *currentObject;
    SRObjectList::Iterator *it = fStarField->NewIterator();
    
    while( (currentObject = it->next()) != 0 )
    {
        currentObject->Draw();
    }
    delete it;
    
    it = fList->NewIterator();
    
    while( (currentObject = it->next()) != 0 )
    {
        currentObject->Draw();
    }
    delete it;
    
    // draw game stats
    char buffer[255];
    RGBColor    textColor = { 0, 0, (UInt32)-1 };
    RGBForeColor( &textColor );
    TextFace(bold);
    
    sprintf( buffer, "Score: %d", fScore );
    
    // draw score
    MoveTo( 8, kGameWindowHeight + kGameStatusBarHeight - 8 );
    DrawText( buffer, 0, strlen(buffer) );
    
    sprintf( buffer, (fTimesDied == 1) ? "Died %d time":"Died %d times", fTimesDied );
    
    // draw times died
    MoveTo( kGameWindowWidth / 2, kGameWindowHeight + kGameStatusBarHeight - 8 );
    DrawText( buffer, 0, strlen(buffer) );
}

void SRGameObject::Think()
{
    SRObject *currentObject;
    SRObjectList::Iterator *it = fStarField->NewIterator();
    Str255 ewhat;
    
    while( (currentObject = it->next()) != 0 )
    {
        currentObject->Think();
    }
    delete it;
    
    it = fList->NewIterator();
    
    while( (currentObject = it->next()) != 0 )
    {
        currentObject->Think();
    }
    
    it->reset();
    
    //dlog->write( "--- frame ---\n", strlen("--- frame ---\n") );
    
    while( (currentObject = it->next()) != 0 )
    {
        SRObjectList *hitList = 0;
        bool wasHit = false;
        
        if(currentObject->WantsCollisions())
        {
            hitList = TestForHit(currentObject);
            if(hitList == 0) continue;
            
            // traverse hit list, telling the currentObject that it hit those
            SRObjectList::Iterator *hit = hitList->NewIterator();
            SRObject *object;
            
            while( (object = hit->next()) != 0 )
            {
                if(currentObject->Hit(object))
                {
                    wasHit = true;
                    break;
                }
            }
            
            // all hit objects must be deleted
            hit->reset();
            
            while( (object = hit->next()) != 0 )
            {
                if(object->Type() != kPlayerType)
                {
                    fList->remove(object);
                    //delete object;
                }
            }
            
            delete hit;
            delete hitList;
            
            // if at least one of the objects caused this one to die,
            // we have to delete it
            if(wasHit && (currentObject->Type() != kPlayerType))
            {
                fList->remove(currentObject);
                //delete currentObject;
            }
        }
    }
    
    delete it;
    fList->clean();
    
    SRSoundManager::MusicIdle();
}

void SRGameObject::Move()
{
    SRObject *currentObject;
    SRObjectList::Iterator *it = fStarField->NewIterator();
    
    while( (currentObject = it->next()) != 0 )
    {
        currentObject->Move();
    }
    delete it;    
    
    it = fList->NewIterator();
    
    while( (currentObject = it->next()) != 0 )
    {
        currentObject->Move();
    }
    delete it;
}

// This is the meat of the collision detection
// The detection is rectangle based, based on the width/height of
// the two objects being considered.
SRObjectList *SRGameObject::TestForHit( SRObject *querant )
{
    //dlog->write( ">\n", 2 );
    char string[255];
    bool hit = false;

    SRObject *currentObject;
    SRObjectList *hitList;
    SRObjectList::Iterator *it;

    hitList = new SRObjectList(false, false);
    
    it = fList->NewIterator();
    
    //char type[5];
    //sprintf( string, "---\nquerant: 0x%x\n", querant );
    //dlog->write(string, strlen(string));
    
    //type[4] = 0;
    //*((int*)type) = querant->Type();
    //sprintf( string, "type:   %s\n", type );
    //dlog->write( string, strlen(string) );

    while( (currentObject = it->next()) != 0 )
    {
        // ignore enemy command objects, they have no physical presence
        if( currentObject->Type() == kEnemyCommandType ||
            currentObject->Type() == kExplosionType ||
            currentObject->Type() == kStarfieldType )
        {
            continue;
        }
    
        if(currentObject != querant)
        {
            // what is the distance between the two objects, in x and y
            int yDiff = querant->yPosition() - currentObject->yPosition();
            int xDiff = querant->xPosition() - currentObject->xPosition();
            
            // these are the minimum distances between the centers of the
            // two sprites
            int heightTotal = (querant->height() + currentObject->height())/2;
            int widthTotal = (querant->width() + currentObject->width())/2;
            
            bool xHit = false, yHit = false;

            // check if the two sprites are vertically close enough
            if( ((yDiff >= 0) && (yDiff <= heightTotal)) ||
                ((yDiff < 0) && (-yDiff <= heightTotal)) )
            {
                yHit = true;
            }

            // check if the two sprites are horizontally close enough
            if( ((xDiff >= 0) && (xDiff <= widthTotal)) ||
                ((xDiff < 0) && (-xDiff <= widthTotal)) )
            {
                xHit = true;
            }
            
            // were the two sprites close enough on both axes?
            if( xHit && yHit )
            {
                hit = hit || true;

                //type[4] = 0;
                //*((int*)type) = currentObject->Type();
                //sprintf( string, "hit by: 0x%x\n", currentObject );
                //dlog->write( string, strlen(string) );
                //sprintf( string, "type:   %s\n", type );
                //dlog->write( string, strlen(string) );
                // yes, then we'll tell the second object it was hit by the querant
                // and add it to our list of victims
                if( currentObject->Hit(querant) )
                    hitList->add(currentObject);
            }
        }
    }
    
    delete it;

    //dlog->write( "<\n", 2 );
    if(hit) return hitList;
    else
    {
        delete hitList;
        return 0;
    }
}

// This function recycles a parent-less object
void SRGameObject::Recycle(SRObject *object)
{
    fList->remove(object);
    //delete object;
}

void SRGameObject::TogglePause()
{
    fPause = !fPause;
    if(fPause)
    {
        SRSoundManager::PauseSoundEffects();
        SRSoundManager::PauseMusic();
    }
    else
    {
        SRSoundManager::ResumeSoundEffects();
        SRSoundManager::ResumeMusic();
    }
}

void SRGameObject::CreateStarField()
{
    fStarField = new SRObjectList( false, true );
    
    int i,j;
    short random;
    float rands[6];
    SRStar *star;
    RGBColor colour;
    
    for( i=0; i<kNumStars; i++ )
    {
        for( j=0; j<6; j++ )
        {
            random = Random();
        
            // invert it
            if(random < 0) random = -random;
        
            // generate a number from 0 to 1.0
            rands[j] = 0;
            if(random != 0) rands[j] = 32767.0 / random;
        }
        
        colour.red = (UInt16) ( 32767 - (rands[0] * kColorShift) );
        colour.green = (UInt16) ( 32767 - (rands[1] * kColorShift) );
        colour.blue = (UInt16) ( 32767 - (rands[2] * kColorShift) );
        
        star = new SRStar( kGameWindowWidth * rands[3], kGameWindowHeight * rands[4], colour );
        star->pushY( kMaxStarSpeed * rands[5] );
        
        fStarField->add(star);
    }
}