#include "SRObjectList.h"
#include "FileWrapper.h"
#include "HFileUtils.h"
#include <stdio.h>

//#include <iostream>
//#include <fstream>
#include <string.h>

// from main.h
extern void DieWithError();

SRObjectList::SRObjectList(bool log, bool fullClean) : fNumItems(0), dlog(0), 
    fCurrentSize(kStartingSize), fFullClean(fullClean)
{
    FSSpec spec = GetCurrentApplicationFSSpec();
    spec.name[0] = 5;
    spec.name[1] = 'd';
    spec.name[2] = 'e';
    spec.name[3] = 'b';
    spec.name[4] = 'u';
    spec.name[5] = 'g';
    if(log)
    {
        dlog = new FileWrapper( &spec, FileWrapper::kAny, true, 'TEXT', 'R*ch' );
        if(!dlog->open()) SysBeep(1);
    }
    
    try{
    fArray = new ListItem[kStartingSize];
    }
    catch(...){
        DieWithError();
    }
/*    char string[255];
    sprintf(string, "0x%x.log", this);
    if(log) fOutFile.open(string);*/
}

SRObjectList::~SRObjectList()
{
    if(dlog)dlog->close();
    
    if(fFullClean)
    {
        int i;
        for( i=0; i<fNumItems; i++ )
        {
            delete (fArray[i].object);
        }
    }
    
    delete [] fArray;
    //fOutFile.close();
/*  Old implementation
    ListItem *currentItem = fFirst;
    ListItem *tempItem;
    
    while( currentItem != 0 )
    {
        tempItem = currentItem;
        currentItem = tempItem->next;
        free(tempItem);
    }
    */
}

SRObjectList::Iterator *SRObjectList::NewIterator()
{
    return new Iterator(fArray, fNumItems);
}

void SRObjectList::printList()
{
/*    if(!dlog) return;
    //fOutFile << "---------------\n";
    dlog->write("---------------\n",strlen("---------------\n"));
    if( fFirst == 0 )
    {
        //fOutFile << "first item dead...\n";
        dlog->write("first item dead...\n",strlen("first item dead...\n"));
        if( fLast != 0 )
            //fOutFile << "but last item living... (?!?!)\n";
            dlog->write("but last item living... (?!?!)\n",
                        strlen("but last item living... (?!?!)\n"));
        return;
    }
    
    char string[255];
    ListItem *currentItem = fFirst;
    do
    {
        sprintf( string, "pointer:     0x%x\n", currentItem );
        //fOutFile << string;
        dlog->write(string,strlen(string));
        
        sprintf( string, "is marked:   %d\n", currentItem->marked );
        dlog->write(string,strlen(string));
        //fOutFile << string;
        
        sprintf( string, "object ptr:  0x%x\n", currentItem->object );
        dlog->write(string,strlen(string));
        dlog->write("---\n",strlen("---\n"));
        //fOutFile <<  string << "---\n";
        
        currentItem = currentItem->next;
    }while(currentItem != 0);*/

}

void SRObjectList::add( SRObject *item )
{
    if( fNumItems == fCurrentSize )
    {
        ListItem *oldArray = fArray;
        
        fCurrentSize += kGrowSize;
        fArray = new ListItem[fCurrentSize];
        
        int i;
        for( i=0; i<fNumItems; i++ )
        {
            fArray[i] = oldArray[i];
        }

        delete [] oldArray;
    }
    
    fArray[fNumItems].object = item;
    fArray[fNumItems++].marked = false;
/*  Old implementation
    if(item == 0) return;
    // when we get a new item, add it to the end of the list
    ListItem *lItem = (ListItem*)malloc(sizeof(ListItem));
    lItem->object = item;
    lItem->next = 0;
    lItem->marked = false;
    
    if(fLast != 0) fLast->next = lItem;
    
    fLast = lItem;
    
    if( fFirst == 0 ) fFirst = lItem;
    
    fNumItems++;
    
    //cout << "Added object " << item << " to list " << this << endl;
    */
}

// called to remove an item from the list
void SRObjectList::remove( SRObject *item )
{
    int i;
    for( i=0; i<fNumItems; i++ )
    {
        if(fArray[i].object == item)
        {
            fArray[i].marked = true;
            break;
        }
    }
/*  Old implementation
    // mandatory sanity checks...
    if(fNumItems < 1) return;
    if(item == 0) return;
    ListItem *currentItem;
    
    currentItem = fFirst;
        
    while( currentItem != 0 )
    {
        if( currentItem->object == item )
        {
            currentItem->marked = true;
            break;
        }
            
        currentItem = currentItem->next;
    }
    */
}

void SRObjectList::clean()
{
    int i, j;
    for( i=0; i<fNumItems; i++ )
    {
        // is this item marked?
        if(fArray[i].marked)
        {
            // yes?  then delete it
            delete (fArray[i].object);

            // one less item
            fNumItems--;

            // and compact the items
            for( j=i; j<fNumItems; j++ )
            {
                fArray[j] = fArray[j+1];
            }
            // and we move back one, since i now points to 
            // the first unprocessed item
            i--;
        }
    }
/*  Old implementation
    // mandatory sanity checks...
    if(fNumItems < 1) return;
    
    //printList();

    ListItem *currentItem, *tempItem, *marked;
    
    while((fFirst != 0) && fFirst->marked)
    {
        currentItem = fFirst;
        fFirst = currentItem->next;
        fNumItems--;
        free(currentItem);
    }
    
    if(fNumItems == 0)
    {
        return;
    }
    
    tempItem = fFirst;
    currentItem = fFirst->next;
    
    while( currentItem != 0 )
    {
        if(currentItem->marked)
        {
            marked = currentItem;
            tempItem->next = currentItem->next;
            currentItem = currentItem->next;
            
            if(marked == fLast)
            {
                fLast = tempItem;
            }
            fNumItems--;
            
            delete marked->object;
            free(marked);
            
            continue;
        }
        tempItem = currentItem;
        currentItem = currentItem->next;
    }
    
    if(fNumItems == 1)
    {
        fFirst = fLast;
        return;
    }

    fLast->next = 0;
    
    //printList();
    */
}

// List iterator
/* Old constructor
SRObjectList::Iterator::Iterator( SRObjectList::ListItem *first ) : fFirst(first),
     fCurrent(first)
     */
SRObjectList::Iterator::Iterator( SRObjectList::ListItem *list, int numItems ) :
    fList(list), fCurrentIndex(0), fNumItems(numItems)
{
}

void SRObjectList::Iterator::reset()
{
    fCurrentIndex = 0;
}

SRObject *SRObjectList::Iterator::next()
{
    if( fCurrentIndex >= fNumItems )
        return 0;
    else
    {
        SRObject *obj = fList[fCurrentIndex].object;
        fCurrentIndex++;
        return obj;
    }
/*  Old implementation
    SRObject *object = 0;
    if(fCurrent != 0)
    {
        object = fCurrent->object;
        fCurrent = fCurrent->next;
    }
        
    return object;
    */
}